
//////////////////////////////////////////////////////////////////////////////
// Includes

#include "fluids/fluids.h"
#include "util/assert.h"

#include <fstream>


//////////////////////////////////////////////////////////////////////////////
// Default Constructor

Fluid::Fluid() : m_timestep(1),
                 m_diffusionCoefficient(0),
                 m_particleBouyancy(0),
                 m_numParticles(0),
                 m_numStreamlines(0),
                 m_numPointsPerStreamline(0),
                 m_currentStreamline(0),
                 m_curSlinePoint(0)
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Virtual Destructor

Fluid::~Fluid()
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// StepSimulation

void Fluid::StepSimulation()
{
    AdvectVorticities();
    AddForces();
    VorticitiesToFluxes();

    // The previous three steps changed the fluxes. Before moving on we have to
    // make sure we keep the velocity field in sync with it.
    FluxesToVelocities();

    if( m_numParticles > 0 )
        AdvectParticles();
}

//////////////////////////////////////////////////////////////////////////////
// InitializeArrays - Resize them and set everything to 0

void Fluid::InitializeVizArrays()
{
    m_particles.resize( m_numParticles );
    m_streamlines.assign( m_numStreamlines, std::vector<vec3f>(m_numPointsPerStreamline, 0.0f) );
    m_slineTangents.assign( m_numStreamlines, std::vector<vec3f>(m_numPointsPerStreamline, 0.0f) );
    m_colors.assign( m_numParticles, 0.0f );
}

//////////////////////////////////////////////////////////////////////////////

void Fluid::BeginStreamline()
{
    m_curSlinePoint = 0;
}

void Fluid::AddStreamlinePoint( const vec3& point )
{
    if( m_curSlinePoint < m_numPointsPerStreamline )
        m_streamlines[m_currentStreamline][m_curSlinePoint++].set_value( (float)point[0], (float)point[1], (float)point[2] );
}

void Fluid::EndStreamline()
{
    smart_assert( m_curSlinePoint > 0 );
    fill( m_streamlines[m_currentStreamline].begin() + m_curSlinePoint,
          m_streamlines[m_currentStreamline].end(), m_streamlines[m_currentStreamline][m_curSlinePoint-1] );

    // the tangent at the first point
    vec3f tangent = m_streamlines[m_currentStreamline][1] - m_streamlines[m_currentStreamline][0];
    tangent.normalize();
    m_slineTangents[m_currentStreamline][0] = tangent;

    // the tangent at the last point
    tangent = m_streamlines[m_currentStreamline][m_curSlinePoint-1] - m_streamlines[m_currentStreamline][m_curSlinePoint-2];
    tangent.normalize();
    m_slineTangents[m_currentStreamline][m_curSlinePoint-1] = tangent;

    // the tangents at the rest of the points
    for( size_t i = 1; i+1 < m_curSlinePoint; ++i )
    {
        tangent = m_streamlines[m_currentStreamline][i+1] - m_streamlines[m_currentStreamline][i-1];
        tangent.normalize();
        m_slineTangents[m_currentStreamline][i] = tangent;
    }

    if( ++m_currentStreamline == m_numStreamlines )
        m_currentStreamline = 0;
}

//////////////////////////////////////////////////////////////////////////////
// DumpParticlesToFile

void Fluid::DumpParticlesToFile( const char *filename )
{
    std::ofstream file( filename, std::ios::out | std::ios::binary );
    size_t numParticles = m_particles.size();
    file.write( reinterpret_cast<char*>(&numParticles), sizeof(size_t) );
    file.write( reinterpret_cast<char*>(&m_particles[0]), 3 * sizeof(float) * numParticles );
}
