
//////////////////////////////////////////////////////////////////////////////
// Includes

#include "fluids/fluids.h"
#include "fluids/matrixBuilder.h"
#include "simplicialComplex/vertex.h"
#include "simplicialComplex/edge.h"
#include "simplicialComplex/iterator.h"
#include "util/filename.h"
#include "util/default_logs.h"

#include <fstream>
#include <string>
using std::string;


//////////////////////////////////////////////////////////////////////////////
// M2SM - Converts MatrixMultiplier objects to solver matrices

static inline void M2SM( const MatrixMultiplier& m, SparseMatrix *sm )
{
    for( size_t i = 0; i < m.rows.size(); ++i )
        for( std::map<size_t,real>::const_iterator elem = m.rows[i].begin(); elem != m.rows[i].end(); ++elem )
            sm->Set( (int)i, (int)elem->first, elem->second );
}

//////////////////////////////////////////////////////////////////////////////
// BuildOrLoadMatrices - Builds the matrices corresponding to the compositions
//                       of the DEC operators that we'll be using, and then
//                       saves them to disk. If the files are already there,
//                       it loads them instead.

void Fluid3D::BuildOrLoadMatrices( bool freeBoundaries, const char *meshFilename )
{
    std::string matFilePrefix = Filename(meshFilename).GetPath() + "matrices/" + Filename(meshFilename).GetName();

    if( freeBoundaries )
        matFilePrefix += "fb";

    // This is the diagonal matrix that kills the rows of the boundary edges
    DiagonalMatrix P(0);
    if( !freeBoundaries )
    {
        P.resize( m_mesh.NumEdges(), 0 );
        for( SimplexIterator<Edge> edge = m_mesh.Edges(); edge; ++edge )
            if( !edge->IsOnBoundary() )
                P[ edge->Index() ] = (real)1.0;
    }

    // d of a primal 1-form
    string filename = matFilePrefix + string("_dp1");
    if( DoesFileExist( Filename(filename) ) )
    {
        LOG(info) << "  Loading d matrix from disk";
        d_P1.Load( filename.c_str() );
    }
    else
    {
        LOG(info) << "  Building d matrix";
        d_P1.BeginConstruction( m_mesh );
        MatrixMultiplier d1 = MatrixBuilder::GetD1( m_mesh );
        M2SM( d1, &d_P1 );
        LOG(info) << "    Constructing preconditioner";
        d_P1.EndConstruction();
        d_P1.Save( filename.c_str() );
    }

    // d* of a primal 2-form
    filename = matFilePrefix + string("_dstarp2");
    if( DoesFileExist( Filename(filename) ) )
    {
        LOG(info) << "  Loading d* matrix from disk";
        d_star_P2.Load( filename.c_str() );
    }
    else
    {
        LOG(info) << "  Building d* matrix";
        d_star_P2.BeginConstruction( m_mesh );
        DiagonalMatrix s2 = MatrixBuilder::GetStar2( m_mesh );
        MatrixMultiplier d1t = MatrixBuilder::GetD1t( m_mesh );
        if( freeBoundaries )
            M2SM( d1t * s2, &d_star_P2 );
        else
            M2SM( P * d1t * s2, &d_star_P2 );
        LOG(info) << "    Constructing preconditioner";
        d_star_P2.EndConstruction();
        d_star_P2.Save( filename.c_str() );
    }

    // (*d*d* + d*d) of a primal 1-form
    filename = matFilePrefix + string("_sdsdspdsdp1");
    if( DoesFileExist( Filename(filename) ) )
    {
        LOG(info) << "  Loading (*d*d* + d*d) matrix from disk";
        sdsdspdsd_P1.Load( filename.c_str() );
    }
    else
    {
        LOG(info) << "  Building (*d*d* + d*d) matrix";

        sdsdspdsd_P1.BeginConstruction( m_mesh, true );

        {
            DiagonalMatrix s0inv = MatrixBuilder::GetStar0( m_mesh );
            s0inv.Invert();

            if( !freeBoundaries )
            {
                // Kill the contributions of the boundary vertices
                for( SimplexIterator<Vertex> vertex = m_mesh.Vertices(); vertex; ++vertex )
                    if( vertex->IsOnBoundary() )
                        s0inv[ vertex->Index() ] = (real)0.0;
            }

            DiagonalMatrix   s1  = MatrixBuilder::GetStar1( m_mesh );
            DiagonalMatrix   s2  = MatrixBuilder::GetStar2( m_mesh );
            MatrixMultiplier d0  = MatrixBuilder::GetD0( m_mesh );
            MatrixMultiplier d0t = MatrixBuilder::GetD0t( m_mesh );
            MatrixMultiplier d1  = MatrixBuilder::GetD1( m_mesh );
            MatrixMultiplier d1t = MatrixBuilder::GetD1t( m_mesh );

            MatrixMultiplier left = (s1 * d0) * s0inv * (d0t * s1);
            MatrixMultiplier right = d1t * s2 * d1;
            left += right;

            if( !freeBoundaries )
            {
                left = P * left * P;
                for( SimplexIterator<Edge> edge = m_mesh.Edges(); edge; ++edge )
                    if( edge->IsOnBoundary() )
                        left.Set( edge->Index(), edge->Index(), (real)1.0 );
            }

            M2SM( left, &sdsdspdsd_P1 );
        }

        LOG(info) << "    Constructing preconditioner";
        sdsdspdsd_P1.EndConstruction();
        sdsdspdsd_P1.Save( filename.c_str() );
    }
}
