#ifndef _FLUIDS_MATRIX_BUILDER_H_
#define _FLUIDS_MATRIX_BUILDER_H_


//////////////////////////////////////////////////////////////////////////////
// Includes

#include "simplicialComplex/simplicialComplex.h"

#include <vector>
#include <map>


//////////////////////////////////////////////////////////////////////////////
// Some really slow, thrown-together, sparse matrix multiplication code

struct MatrixMultiplier
{
    MatrixMultiplier( size_t numRows, size_t numColumns );

    void Set( size_t i, size_t j, real val );
    void Add( size_t i, size_t j, real val );

    void operator+=( const MatrixMultiplier& mat );

    size_t numRows, numColumns;
    std::vector< std::map<size_t,real> > rows;
};

struct DiagonalMatrix : public std::vector<real>
{
    DiagonalMatrix( size_t size ) : std::vector<real>(size) {}
    void Invert();
};

MatrixMultiplier operator*( const MatrixMultiplier& src1, const MatrixMultiplier& src2 );
MatrixMultiplier operator*( const MatrixMultiplier& src1, const DiagonalMatrix& src2 );
MatrixMultiplier operator*( const DiagonalMatrix& src1, const MatrixMultiplier& src2 );


//////////////////////////////////////////////////////////////////////////////
// class MatrixBuilder
// Builds the DEC matrices for a SimplicialComplex object.

class MatrixBuilder
{
public:
    // These functions convert the adjacency list representation into a
    // sparse matrix representation, so that the matrices can be multiplied
    template<int N> static MatrixMultiplier GetD0( const SimplicialComplex<N>& mesh );
    template<int N> static MatrixMultiplier GetD1( const SimplicialComplex<N>& mesh );
    template<int N> static MatrixMultiplier GetD2( const SimplicialComplex<N>& mesh );

    template<int N> static MatrixMultiplier GetD0t( const SimplicialComplex<N>& mesh );
    template<int N> static MatrixMultiplier GetD1t( const SimplicialComplex<N>& mesh );
    template<int N> static MatrixMultiplier GetD2t( const SimplicialComplex<N>& mesh );

    // The Hodge star matrices
    template<int N> static DiagonalMatrix GetStar0( const SimplicialComplex<N>& mesh );
    template<int N> static DiagonalMatrix GetStar1( const SimplicialComplex<N>& mesh );
    template<int N> static DiagonalMatrix GetStar2( const SimplicialComplex<N>& mesh );
    template<int N> static DiagonalMatrix GetStar3( const SimplicialComplex<N>& mesh );
};


//////////////////////////////////////////////////////////////////////////////
template<int N> MatrixMultiplier MatrixBuilder::GetD0( const SimplicialComplex<N>& sc )
{
    MatrixMultiplier D0( sc.NumEdges(), sc.NumVerts() );
    for( size_t edge = 0; edge < sc.NumEdges(); ++edge )
    {
        D0.Set( edge, sc.m_EVmatrix[edge][0], (real)(-1.0) );
        D0.Set( edge, sc.m_EVmatrix[edge][1], (real)( 1.0) );
    }
    return D0;
}

//////////////////////////////////////////////////////////////////////////////
template<int N> MatrixMultiplier MatrixBuilder::GetD1( const SimplicialComplex<N>& sc )
{
    MatrixMultiplier D1( sc.NumFaces(), sc.NumEdges() );
    for( size_t face = 0; face < sc.NumFaces(); ++face )
    {
        D1.Set( face, sc.m_FEmatrix[face][0].first, (real)sc.m_FEmatrix[face][0].second );
        D1.Set( face, sc.m_FEmatrix[face][1].first, (real)sc.m_FEmatrix[face][1].second );
        D1.Set( face, sc.m_FEmatrix[face][2].first, (real)sc.m_FEmatrix[face][2].second );
    }
    return D1;
}

//////////////////////////////////////////////////////////////////////////////
template<int N> MatrixMultiplier MatrixBuilder::GetD2( const SimplicialComplex<N>& sc )
{
    MatrixMultiplier D2( sc.NumTets(), sc.NumFaces() );
    for( size_t tet = 0; tet < sc.NumTets(); ++tet )
    {
        D2.Set( tet, sc.m_TFmatrix[tet][0].first, (real)sc.m_TFmatrix[tet][0].second );
        D2.Set( tet, sc.m_TFmatrix[tet][1].first, (real)sc.m_TFmatrix[tet][1].second );
        D2.Set( tet, sc.m_TFmatrix[tet][2].first, (real)sc.m_TFmatrix[tet][2].second );
        D2.Set( tet, sc.m_TFmatrix[tet][3].first, (real)sc.m_TFmatrix[tet][3].second );
    }
    return D2;
}

//////////////////////////////////////////////////////////////////////////////
template<int N> MatrixMultiplier MatrixBuilder::GetD0t( const SimplicialComplex<N>& sc )
{
    MatrixMultiplier D0t( sc.NumVerts(), sc.NumEdges() );
    for( size_t edge = 0; edge < sc.NumEdges(); ++edge )
    {
        D0t.Set( sc.m_EVmatrix[edge][0], edge, (real)(-1.0) );
        D0t.Set( sc.m_EVmatrix[edge][1], edge, (real)( 1.0) );
    }
    return D0t;
}

//////////////////////////////////////////////////////////////////////////////
template<int N> MatrixMultiplier MatrixBuilder::GetD1t( const SimplicialComplex<N>& sc )
{
    MatrixMultiplier D1t( sc.NumEdges(), sc.NumFaces() );
    for( size_t face = 0; face < sc.NumFaces(); ++face )
    {
        D1t.Set( sc.m_FEmatrix[face][0].first, face, (real)sc.m_FEmatrix[face][0].second );
        D1t.Set( sc.m_FEmatrix[face][1].first, face, (real)sc.m_FEmatrix[face][1].second );
        D1t.Set( sc.m_FEmatrix[face][2].first, face, (real)sc.m_FEmatrix[face][2].second );
    }
    return D1t;
}

//////////////////////////////////////////////////////////////////////////////
template<int N> MatrixMultiplier MatrixBuilder::GetD2t( const SimplicialComplex<N>& sc )
{
    MatrixMultiplier D2t( sc.NumFaces(), sc.NumTets() );
    for( size_t tet = 0; tet < sc.NumTets(); ++tet )
    {
        D2t.Set( sc.m_TFmatrix[tet][0].first, tet, (real)sc.m_TFmatrix[tet][0].second );
        D2t.Set( sc.m_TFmatrix[tet][1].first, tet, (real)sc.m_TFmatrix[tet][1].second );
        D2t.Set( sc.m_TFmatrix[tet][2].first, tet, (real)sc.m_TFmatrix[tet][2].second );
        D2t.Set( sc.m_TFmatrix[tet][3].first, tet, (real)sc.m_TFmatrix[tet][3].second );
    }
    return D2t;
}

//////////////////////////////////////////////////////////////////////////////
template<int N> DiagonalMatrix MatrixBuilder::GetStar0( const SimplicialComplex<N>& sc )
{
    DiagonalMatrix star( sc.NumVerts() );
    for( std::vector< Simplex<0> >::const_iterator simplex = sc.m_vertices.begin(); simplex != sc.m_vertices.end(); ++simplex )
        star[ simplex->Index() ] = simplex->DualVolume() / simplex->Volume();
    return star;
}

//////////////////////////////////////////////////////////////////////////////
template<int N> DiagonalMatrix MatrixBuilder::GetStar1( const SimplicialComplex<N>& sc )
{
    DiagonalMatrix star( sc.NumEdges() );
    for( std::vector< Simplex<1> >::const_iterator simplex = sc.m_edges.begin(); simplex != sc.m_edges.end(); ++simplex )
        star[ simplex->Index() ] = simplex->DualVolume() / simplex->Volume();
    return star;
}

//////////////////////////////////////////////////////////////////////////////
template<int N> DiagonalMatrix MatrixBuilder::GetStar2( const SimplicialComplex<N>& sc )
{
    DiagonalMatrix star( sc.NumFaces() );
    for( std::vector< Simplex<2> >::const_iterator simplex = sc.m_faces.begin(); simplex != sc.m_faces.end(); ++simplex )
        star[ simplex->Index() ] = simplex->DualVolume() / simplex->Volume();
    return star;
}

//////////////////////////////////////////////////////////////////////////////
template<int N> DiagonalMatrix MatrixBuilder::GetStar3( const SimplicialComplex<N>& sc )
{
    DiagonalMatrix star( sc.NumTets() );
    for( std::vector< Simplex<3> >::const_iterator simplex = sc.m_tets.begin(); simplex != sc.m_tets.end(); ++simplex )
        star[ simplex->Index() ] = simplex->DualVolume() / simplex->Volume();
    return star;
}


#endif  // _FLUIDS_MATRIX_BUILDER_H_
