
//////////////////////////////////////////////////////////////////////////////
// Includes

#include "simplicialComplex/face.h"
#include "simplicialComplex/simplicialComplex.h"
#include "simplicialComplex/vertex.h"
#include "simplicialComplex/edge.h"
#include "simplicialComplex/tet.h"
#include "util/assert.h"


//////////////////////////////////////////////////////////////////////////////
// Default Constructor

Face::Face() : sc(0), index(size_t(-1))
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Private Constructor that takes a mesh pointer and index

Face::Face( const SimplicialComplex<3> *sc, size_t index ) : sc(sc), index(index)
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

Face::~Face()
{
    // empty
}

//////////////////////////////////////////////////////////////////////////////
// operator bool - Is this face valid

Face::operator bool() const
{
    return (sc && index < sc->NumFaces());
}

//////////////////////////////////////////////////////////////////////////////
// Center

vec3 Face::Center() const
{
    smart_assert(*this).msg( "Invalid face" ).fatal();
    return sc->m_faces[index].Center();
}

//////////////////////////////////////////////////////////////////////////////
// Area

real Face::Area() const
{
    smart_assert(*this).msg( "Invalid face" ).fatal();
    return sc->m_faces[index].Volume();
}

//////////////////////////////////////////////////////////////////////////////
// DualLength

real Face::DualLength() const
{
    smart_assert(*this).msg( "Invalid face" ).fatal();
    return sc->m_faces[index].DualVolume();
}

//////////////////////////////////////////////////////////////////////////////
// Normal

vec3 Face::Normal() const
{
    smart_assert(*this).msg( "Invalid face" ).fatal();
    vec3 v0, v1, v2;
    Vertices( v0, v1, v2 );
    vec3 normal = vec3(v1 - v0).cross(v2 - v0);
    normal.normalize();
    return normal;
}

//////////////////////////////////////////////////////////////////////////////
// Vertices

void Face::Vertices( vec3& v0, vec3& v1, vec3& v2 ) const
{
    smart_assert(*this).msg( "Invalid face" ).fatal();
    v0 = sc->m_vertices[ sc->m_faces[index].Get(0) ].Center();
    v1 = sc->m_vertices[ sc->m_faces[index].Get(1) ].Center();
    v2 = sc->m_vertices[ sc->m_faces[index].Get(2) ].Center();
}

//////////////////////////////////////////////////////////////////////////////
// Vertices

void Face::Vertices( Vertex& v0, Vertex& v1, Vertex& v2 ) const
{
    smart_assert(*this).msg( "Invalid face" ).fatal();
    v0.sc = v1.sc = v2.sc = sc;
    v0.index = sc->m_faces[index].Get(0);
    v1.index = sc->m_faces[index].Get(1);
    v2.index = sc->m_faces[index].Get(2);
}

//////////////////////////////////////////////////////////////////////////////
// Edges

void Face::Edges( Edge& e0, Edge& e1, Edge& e2 ) const
{
    smart_assert(*this).msg( "Invalid face" ).fatal();
    e0.sc = e1.sc = e2.sc = sc;
    e0.index = sc->m_FEmatrix[ index ][ 0 ].first;
    e1.index = sc->m_FEmatrix[ index ][ 1 ].first;
    e2.index = sc->m_FEmatrix[ index ][ 2 ].first;
}

//////////////////////////////////////////////////////////////////////////////
// Tets

void Face::Tets( Tet& t0, Tet& t1 ) const
{
    smart_assert(*this).msg( "Invalid face" ).fatal();
    t0.sc = t1.sc = sc;
    t0.index = sc->m_FTmatrix[ index ][ 0 ];
    t1.index = sc->m_FTmatrix[ index ][ 1 ];
}

//////////////////////////////////////////////////////////////////////////////
// IsOnBoundary

bool Face::IsOnBoundary() const
{
    smart_assert(*this).msg( "Invalid face" ).fatal();
    // It's a boundary face if one of its two tets is nonexistant
    return ( sc->m_FTmatrix[index][0] == size_t(-1) || sc->m_FTmatrix[index][1] == size_t(-1) );
}

//////////////////////////////////////////////////////////////////////////////
// Equality operator

bool operator==( const Face& lhs, const Face& rhs )
{
    return (lhs.sc == rhs.sc && lhs.index == rhs.index);
}

//////////////////////////////////////////////////////////////////////////////
// Inequality operator

bool operator!=( const Face& lhs, const Face& rhs )
{
    return (!(lhs == rhs));
}
