
//////////////////////////////////////////////////////////////////////////////
// Includes

#include "util/color.h"


//////////////////////////////////////////////////////////////////////////////
// Color space conversion

vec3f RGBtoHSV( const vec3f& rgb )
{
    const float R = rgb[0], G = rgb[1], B = rgb[2];
    float H = -1.0f, S, V;

    // RGB_TO_HSV from Foley, van Dam, Feiner, Hughes 2nd Ed. Pg 592
    const float maximum = max( R, G, B );
    const float minimum = min( R, G, B );
    const float diff = maximum - minimum;

    V = maximum;

    // Next Saturation
    if( maximum != 0.0f ) S = diff / maximum;
    else                  S = 0.0f;

    // Next Hue
    if( S != 0.0f )
    {
        const float r_dist = ( maximum - R ) / diff;
        const float g_dist = ( maximum - G ) / diff;
        const float b_dist = ( maximum - B ) / diff;

        if      ( R == maximum ) H = b_dist - g_dist;
        else if ( G == maximum ) H = 2.0f + r_dist - b_dist;
        else if ( B == maximum ) H = 4.0f + g_dist - r_dist;

        H /= 6.0f;
        if( H < 0.0f ) H += 1.0f;
    }

    return vec3f( H, S, V );
}

//////////////////////////////////////////////////////////////////////////////
// Color space conversion

vec3f RGBtoHSL( const vec3f& rgb )
{
    const float R = rgb[0], G = rgb[1], B = rgb[2];
    float H = -1.0f, S = 0.0f, L;

    // RGB_TO_HLS from Foley, van Dam, Feiner, Hughes 2nd Ed. Pg 595
    const float maximum = max( R, G, B );
    const float minimum = min( R, G, B );
    const float diff = maximum - minimum;

    L = ( maximum + minimum ) / 2.0f;

    // Next Saturation
    if( diff <= 1e-6f )
        L = 0.0f;
    else
    {
        if( L < 0.5f ) S = diff / ( maximum + minimum );
        else           S = diff / ( 2.0f - maximum - minimum );

        const float r_dist = ( maximum - R ) / diff;
        const float g_dist = ( maximum - G ) / diff;
        const float b_dist = ( maximum - B ) / diff;

        // Next Hue
        if      ( R == maximum ) H = b_dist - g_dist;
        else if ( G == maximum ) H = 2.0f + r_dist - b_dist;
        else if ( B == maximum ) H = 4.0f + g_dist - r_dist;

        H /= 6.0f;
        if( H < 0.0f ) H += 1.0f;
    }

    return vec3f( H, S, L );
}

//////////////////////////////////////////////////////////////////////////////
// Color space conversion

vec3f RGBtoYIQ( const vec3f& rgb )
{
    static matrix4 matRGBtoYIQ( 0.299f,  0.587f,  0.114f, 0.0f,
                                0.596f, -0.274f, -0.322f, 0.0f,
                                0.212f, -0.523f,  0.311f, 0.0f,
                                0.0f,    0.0f,    0.0f,   1.0f );
    vec3f yiq;
    matRGBtoYIQ.mult_matrix_vec( rgb, yiq );
    return yiq;
}

//////////////////////////////////////////////////////////////////////////////
// Color space conversion

vec3f HSVtoRGB( const vec3f& hsv )
{
    float H = hsv[0], S = hsv[1], V = hsv[2];
    float R = 0.0f, G = 0.0f, B = 0.0f;

    // HSV_TO_RGB from Foley, van Dam, Feiner, Hughes 2nd Ed. Pg 593
    if( S == 0.0f )
        R = G = B = V;
    else
    {
        if( H == 1.0f ) H = 0.0f;
        H *= 6.0f;     // H is now in [0,6)
        int i = static_cast<int>( floor(H) );
        float f = H - i;
        float p = V * ( 1.0f - S );
        float q = V * ( 1.0f - ( S * f ) );
        float t = V * ( 1.0f - ( S * ( 1.0f - f ) ) );
        switch ( i )
        {
        case 0: R = V; G = t; B = p; break;
        case 1: R = q; G = V; B = p; break;
        case 2: R = p; G = V; B = t; break;
        case 3: R = p; G = q; B = V; break;
        case 4: R = t; G = p; B = V; break;
        case 5: R = V; G = p; B = q; break;
        }
    }

    return vec3f( R, G, B );
}

//////////////////////////////////////////////////////////////////////////////
// Color space conversion

static float HSLValue( float n1, float n2, float hue )
{
    float value;

    if( hue > 360 )    hue -= 360;
    else if( hue < 0 ) hue += 360;

    if( hue < 60 )       value = n1 + ( n2 - n1 ) * hue / 60;
    else if( hue < 180 ) value = n2;
    else if( hue < 240 ) value = n1 + ( n2 - n1 ) * ( 240 - hue ) / 60;
    else                 value = n1;

    return value;
}

//////////////////////////////////////////////////////////////////////////////
// Color space conversion

vec3f HSLtoRGB( const vec3f& hsl )
{
    const float H = hsl[0] * 360.0f, S = hsl[1], L = hsl[2];
    float R, G, B;

    // HSL_TO_RGB from Foley, van Dam, Feiner, Hughes 2nd Ed. Pg 596

    float m2, m1;
    if( L <= 0.5f ) m2 = L * ( 1.0f + S );
    else            m2 = L + S - ( L * S );
    m1 = 2 * L - m2;

    if( S == 0.0f )
        R = G = B = L;
    else
    {
        R = HSLValue( m1, m2, H + 120 );
        G = HSLValue( m1, m2, H );
        B = HSLValue( m1, m2, H - 120 );
    }

    return vec3f( R, G, B );
}

//////////////////////////////////////////////////////////////////////////////
// Color space conversion

vec3f YIQtoRGB( const vec3f& yiq )
{
    static matrix4 matRGBtoYIQ( 1.0f,  0.956f,  0.621f, 0.0f,
                                1.0f, -0.272f, -0.647f, 0.0f,
                                1.0f, -1.105f,  1.702f, 0.0f,
                                0.0f,  0.0f,    0.0f,   1.0f );
    vec3f rgb;
    matRGBtoYIQ.mult_matrix_vec( yiq, rgb );
    return rgb;
}

//////////////////////////////////////////////////////////////////////////////
// Color space conversion

vec3f HueToRGB( float H )
{
    // HSVtoRGB with S and V set to 1
    float R = 0.0f, G = 0.0f, B = 0.0f;

    if( H < 0.0f || H >= 1.0f ) H = 0.0f;
    H *= 6.0f;     // H is now in [0,6)
    int i = static_cast<int>( floor(H) );
    float f = H - i;
    switch( i )
    {
    case 0: R = 1;   G = f;   B = 0;   break;
    case 1: R = 1-f; G = 1;   B = 0;   break;
    case 2: R = 0;   G = 1;   B = f;   break;
    case 3: R = 0;   G = 1-f; B = 1;   break;
    case 4: R = f;   G = 0;   B = 1;   break;
    case 5: R = 1;   G = 0;   B = 1-f; break;
    }

    return vec3f( R, G, B );
}