#ifndef _FILENAME_H_
#define _FILENAME_H_

#include <locale>
#include <fstream>
#include <string>

/**
* @class Filename
* @ingroup util
*
* This class describes a file name, formatted for the particular architecture.
* A file name consists of an optional <i>path</i> string followed by the
* <i>file</i>.  The <i>file</i> is composed of a <i>name</i> string followed
* by an optional <i>extension</i> string that is separated by a decimal point.
*/
class Filename {
public:  // methods
    /**
    * Default constructor.  This filename is empty.
    */
    Filename();

    /**
    * Constructs a filename with the specified string.
    * @param filename  filename
    */
    Filename(const std::string& filename);

    /**
    * Destructor.
    */
    ~Filename();

    /**
    * Assignment operator.  This filename is assigned the specified string.
    * @param filename  filename
    */
    Filename& operator=(const std::string& filename);

    /**
    * Returns true if this filename is identical to the specified filename.
    * @param filename  filename to compare
    */
    bool operator==(const Filename& filename) const;

    /**
    * Returns a string representing this filename.
    * @return string representing this filename
    */
    operator std::string() const;

    /**
    * Returns the extension for this filename, in lower case.
    * @return extension for this filename
    */
    std::string GetExtension() const;

    /**
    * Returns the file for this filename.
    * @return file for this filename
    */
    std::string GetFile() const;

    /**
    * Returns the name for this filename.
    * @return name for this filename
    */
    std::string GetName() const;

    /**
    * Returns the path for this filename.
    * @return path for this filename
    */
    std::string GetPath() const;

    /**
    * Returns true if this filename is empty.
    * @return true if this filename is empty
    */
    bool IsEmpty() const;

    /**
    * Sets the extension for this filename.
    * @param extension  extension
    */
    void SetExtension(const std::string& extension);

    /**
    * Sets the file for this filename.
    * @param file  file
    */
    void SetFile(const std::string& file);

    /**
    * Sets the name for this filename.
    * @param name  name
    */
    void SetName(const std::string& name);

    /**
    * Sets the path for this filename.
    * @param path  path
    */
    void SetPath(const std::string& path);

public:  // static state
    /** The architecture-specific path separator character. */
    static const char PATH_SEPARATOR;

private:  // state
    std::string _filename;
};

/** @addtogroup util */
// @{

/**
* Returns true if the specified file exists.
* @param filename  filename
* @return true if <i>filename</i> exists
*/
bool DoesFileExist(const Filename& filename);

/**
* Prints the specified filename to an output stream.
* @param out  output stream
* @param filename  filename
* @return output stream
*/
std::ostream& operator<<(std::ostream& out, const Filename& filename);

// @}

inline Filename::Filename() : _filename() { }

inline Filename::Filename(const std::string& filename) {
    *this = filename;
}

inline Filename::~Filename() { }

inline Filename& Filename::operator=(const std::string& filename) {
    _filename = filename;
    for (unsigned int i = 0; i < _filename.length(); ++i) {
        if (_filename[i] == '/' || _filename[i] == '\\')
            _filename[i] = PATH_SEPARATOR;
    }

    return (*this);
}

inline bool Filename::operator==(const Filename& filename) const {
    return (_filename == filename._filename);
}

inline Filename::operator std::string() const {
    return (GetPath() + GetFile());
}

inline std::string Filename::GetExtension() const {
    std::string extension;

    size_t decimal = _filename.find_last_of('.');
    size_t separator = _filename.find_last_of(PATH_SEPARATOR);
    if (decimal == std::string::npos || (separator != std::string::npos && decimal < separator) )
        return (extension);

    extension.assign(_filename, decimal + 1, _filename.length() - decimal - 1);
    std::use_facet<std::ctype<char> >( std::locale("") ).tolower(
        &extension[0], &extension[extension.length()-1] );

    return (extension);
}

inline std::string Filename::GetFile() const {
    std::string file;

    size_t separator = _filename.find_last_of(PATH_SEPARATOR);
    if (separator == std::string::npos)
        file = _filename;
    else
        file.assign(_filename, separator + 1, _filename.length() - separator - 1);

    return (file);
}

inline std::string Filename::GetName() const {
    std::string name;

    size_t decimal = _filename.find_last_of('.');
    size_t separator = _filename.find_last_of(PATH_SEPARATOR);
    if (decimal == std::string::npos || (separator != std::string::npos && decimal < separator) )
        name.assign(_filename, separator + 1, _filename.length() - separator - 1);
    else
        name.assign(_filename, separator + 1, decimal - separator - 1);

    return (name);
}

inline std::string Filename::GetPath() const {
    std::string path;

    size_t separator = _filename.find_last_of(PATH_SEPARATOR);
    if (separator != std::string::npos)
        path.assign(_filename, 0, separator + 1);

    return (path);
}

inline bool Filename::IsEmpty() const {
    return (_filename.length() == 0);
}

inline void Filename::SetExtension(const std::string& extension) {
    size_t decimal = _filename.find_last_of('.');
    size_t separator = _filename.find_last_of(PATH_SEPARATOR);
    if (decimal == std::string::npos || (separator != std::string::npos && decimal < separator) ) {
        if (!extension.empty())
            _filename += '.' + extension;
    } else {
        if (!extension.empty())
            _filename.replace(decimal + 1, _filename.length() - decimal - 1, extension);
        else
            _filename.erase(decimal, _filename.length() - decimal);
    }
}

inline void Filename::SetFile(const std::string& file) {
    size_t separator = _filename.find_last_of(PATH_SEPARATOR);
    _filename.replace(separator + 1, _filename.length() - separator - 1, file);
}

inline void Filename::SetName(const std::string& name) {
    size_t decimal = _filename.find_last_of('.');
    size_t separator = _filename.find_last_of(PATH_SEPARATOR);
    if (decimal == std::string::npos || (separator != std::string::npos && decimal < separator) )
        _filename.replace(separator + 1, _filename.length() - separator - 1, name);
    else
        _filename.replace(separator + 1, decimal - separator - 1, name);
}

inline void Filename::SetPath(const std::string& path) {
    size_t separator = _filename.find_last_of(PATH_SEPARATOR);
    if (separator == std::string::npos)
        _filename = path + _filename;
    else
        _filename.replace(0, separator + 1, path);

    if (path.empty())
        return;

    for (unsigned int i = 0; i < path.length(); ++i) {
        if (_filename[i] == '/' || _filename[i] == '\\')
            _filename[i] = PATH_SEPARATOR;
    }

    if (_filename[path.length() - 1] != PATH_SEPARATOR)
        _filename.insert(path.length(), &PATH_SEPARATOR, 1);
}

inline bool DoesFileExist(const Filename& filename) {
    std::ifstream in(std::string(filename).c_str());
    return (in.good());
}

inline std::ostream& operator<<(std::ostream& out, const Filename& filename) {
    return (out << std::string(filename));
}

#endif  // _FILENAME_H_
