
//////////////////////////////////////////////////////////////////////////////
// Includes

#include "util/geometry.h"


//////////////////////////////////////////////////////////////////////////////
// Circumcenter - 2D

vec2 Circumcenter( const vec2& v1, const vec2& v2, const vec2& v3 )
{
    vec3 center = Circumcenter( vec3(v1[0],v1[1],0), vec3(v2[0],v2[1],0), vec3(v3[0],v3[1],0) );
    return vec2( center[0], center[1] );
}

//////////////////////////////////////////////////////////////////////////////
// Circumcenter - 3D

vec3 Circumcenter( const vec3& v1, const vec3& v2, const vec3& v3 )
{
    const vec3 e1 = v1 - v3;
    const vec3 e2 = v2 - v3;

    const real e1e1 = e1.dot(e1);
    const real e1e2 = e1.dot(e2);
    const real e2e2 = e2.dot(e2);

    vec2 u( e2e2 * (e1e1 - e1e2), e1e1 * (e2e2 - e1e2) );
    u *= (real)0.5 / (e1e1 * e2e2 - e1e2 * e1e2);

    return (v3 + u[0]*e1 + u[1]*e2);
}

//////////////////////////////////////////////////////////////////////////////
// Circumcenter - 3D

vec3 Circumcenter( const vec3& v1, const vec3& v2, const vec3& v3, const vec3& v4 )
{
    const vec3 e1 = v2 - v1;
    const vec3 e2 = v3 - v1;
    const vec3 e3 = v4 - v1;

    const real e1e1 = e1.square_norm();
    const real e2e2 = e2.square_norm();
    const real e3e3 = e3.square_norm();

    const vec3 e2xe3 = e2.cross(e3);
    const vec3 e3xe1 = e3.cross(e1);
    const vec3 e1xe2 = e1.cross(e2);

    return (v1 + (real)0.5 * (e1e1 * e2xe3 + e2e2 * e3xe1 + e3e3 * e1xe2) / e1.dot(e2xe3));
}

//////////////////////////////////////////////////////////////////////////////
// Barycenter - 2D

vec2 Barycenter( const vec2& v1, const vec2& v2, const vec2& v3 )
{
    return (real)(1.0/3.0) * ( v1 + v2 + v3 );
}

//////////////////////////////////////////////////////////////////////////////
// Barycenter - 3D

vec3 Barycenter( const vec3& v1, const vec3& v2, const vec3& v3, const vec3& v4 )
{
    return (real)0.25 * ( v1 + v2 + v3 + v4 );
}

//////////////////////////////////////////////////////////////////////////////
// Length - of a line segment in 2D

real Length( const vec2& v1, const vec2& v2 )
{
    return (v2-v1).length();
}

//////////////////////////////////////////////////////////////////////////////
// Length - of a line segment in 3D

real Length( const vec3& v1, const vec3& v2 )
{
    return (v2-v1).length();
}

//////////////////////////////////////////////////////////////////////////////
// Area - of a triangle in 2D

real Area( const vec2& v1, const vec2& v2, const vec2& v3 )
{
    const vec2 e1 = v2 - v1;
    const vec2 e2 = v3 - v1;
    return (real)0.5 * ( e1.cross(e2) );
}

//////////////////////////////////////////////////////////////////////////////
// Area - of a triangle in 3D

real Area( const vec3& v1, const vec3& v2, const vec3& v3 )
{
    const vec3 e1 = v2 - v1;
    const vec3 e2 = v3 - v1;
    return (real)0.5 * ( e1.cross(e2) ).length();
}

//////////////////////////////////////////////////////////////////////////////
// Volume - of a tetrahedron in 3D

real Volume( const vec3& v1, const vec3& v2, const vec3& v3, const vec3& v4 )
{
    const vec3 e1 = v2 - v1;
    const vec3 e2 = v3 - v1;
    const vec3 e3 = v4 - v1;

    return ((real)1.0/(real)6.0) * e1.cross( e2 ).dot( e3 );
}

//////////////////////////////////////////////////////////////////////////////
// Cotangent - at vertex v1 of the triangle

real Cotangent( const vec2& v1, const vec2& v2, const vec2& v3 )
{
    const vec2 e1 = v2 - v1;
    const vec2 e2 = v3 - v1;
    return (e1.dot(e2)) / (e1[0] * e2[1] - e1[1] * e2[0] );
}

//////////////////////////////////////////////////////////////////////////////
// PolygonArea - for arbitrary simple polygons

real PolygonArea( const std::vector<vec2>& verts )
{
    if( verts.size() < 3 ) return 0.0;
    int last = (int)verts.size() - 1;
    real area = verts[last][0] * ( verts[0][1] - verts[last-1][1] );
    for( int i0 = last, i1 = 0, i2 = 1; i1 < last; i0 = i1, i1 = i2++ )
        area += verts[i1][0] * ( verts[i2][1] - verts[i0][1] );
    area *= 0.5;
    return area;
}

//////////////////////////////////////////////////////////////////////////////
// PolygonArea - for polygons in 3D

real PolygonArea( std::vector<vec3> verts )
{
    if( verts.size() < 3 )
        return 0.0;

    // Split the (possibly non-planar) polygon up into triangles

    vec3 centroid(0,0,0);
    for( size_t i = 0; i < verts.size(); ++i )
        centroid += verts[i];
    centroid /= (real)verts.size();

    vec3 ringPoint1, ringPoint2;
    vec3 triEdge1, triEdge2;
    vec3 normal;
    real area = 0.0;

    ringPoint2 = verts.back();
    for( size_t j = 0; j < verts.size(); ++j )
    {
        ringPoint1 = ringPoint2;
        ringPoint2 = verts[j];

        triEdge1 = ringPoint1 - centroid;
        triEdge2 = ringPoint2 - centroid;
        normal = triEdge1.cross( triEdge2 );
        area += normal.length();
    }

    area *= 0.5;
    return area;
}

//////////////////////////////////////////////////////////////////////////////
// IsPointInTet

bool IsPointInTet( const vec3& pt, const vec3& v1, const vec3& v2, const vec3& v3, const vec3& v4 )
{
    vec4 bary = BarycentricCoords( pt, v1, v2, v3, v4 );
    if( bary[0] < 0 || bary[1] < 0 || bary[2] < 0 || bary[3] < 0 )
        return false;

    //vec3 diff[4] = { v1 - v4, v2 - v4, v3 - v4, pt - v4 };

    //real M00 = diff[0].dot(diff[0]);
    //real M01 = diff[0].dot(diff[1]);
    //real M02 = diff[0].dot(diff[2]);
    //real M11 = diff[1].dot(diff[1]);
    //real M12 = diff[1].dot(diff[2]);
    //real M22 = diff[2].dot(diff[2]);
    //real R0  = diff[0].dot(diff[3]);
    //real R1  = diff[1].dot(diff[3]);
    //real R2  = diff[2].dot(diff[3]);

    //real C00 = M11*M22-M12*M12;
    //real C01 = M02*M12-M01*M22;
    //real C02 = M01*M12-M02*M11;
    //real C11 = M00*M22-M02*M02;
    //real C12 = M02*M01-M00*M12;
    //real C22 = M00*M11-M01*M01;

    //real denom = M00*C00 + M01*C01 + M02*C02;

    //real numer[4];
    //numer[0] = C00*R0+C01*R1+C02*R2;
    //numer[1] = C01*R0+C11*R1+C12*R2;
    //numer[2] = C02*R0+C12*R1+C22*R2;
    //numer[3] = denom-numer[0]-numer[1]-numer[2];

    //if( denom > 0.0f )
    //{
    //    for( int i = 0; i < 4; ++i )
    //        if( numer[i] < 0.0f || numer[i] > denom )
    //            return false;
    //}
    //else
    //{
    //    for( int i = 0; i < 4; ++i )
    //        if( numer[i] < denom || numer[i] > 0.0f )
    //            return false;
    //}

    return true;
}

//////////////////////////////////////////////////////////////////////////////
// BarycentricCoords - for triangles

vec3 BarycentricCoords( const vec2& pt, const vec2& v1, const vec2& v2, const vec2& v3 )
{
    vec2 e1 = v1 - pt;
    vec2 e2 = v2 - pt;
    vec2 e3 = v3 - pt;

    vec3 baryCoords( e2.cross(e3), e3.cross(e1), e1.cross(e2) );
    real total = baryCoords[0] + baryCoords[1] + baryCoords[2];
    baryCoords /= total;

    return baryCoords;
}

//////////////////////////////////////////////////////////////////////////////
// BarycentricCoords - for tetrahedra

vec4 BarycentricCoords( const vec3& pt, const vec3& v1, const vec3& v2, const vec3& v3, const vec3& v4 )
{
    vec3 e1 = v1 - pt;
    vec3 e2 = v2 - pt;
    vec3 e3 = v3 - pt;
    vec3 e4 = v4 - pt;

    vec4 baryCoords( e2.cross(e3).dot(e4), e3.cross(e1).dot(e4), e4.cross(e1).dot(e2), e1.cross(e3).dot(e2) );
    real total = baryCoords[0] + baryCoords[1] + baryCoords[2] + baryCoords[3];
    baryCoords /= total;

    return baryCoords;
}

//////////////////////////////////////////////////////////////////////////////
// Affine - affine combination of three points

vec3 Affine( const vec3& weights, const vec3& p1, const vec3& p2, const vec3& p3 )
{
    return weights[0] * p1 + weights[1] * p2 + weights[2] * p3;
}

//////////////////////////////////////////////////////////////////////////////
// Affine - affine combination of four points

vec3 Affine( const vec4& weights, const vec3& p1, const vec3& p2, const vec3& p3, const vec3& p4 )
{
    return weights[0] * p1 + weights[1] * p2 + weights[2] * p3 + weights[3] * p4;
}
