#ifndef _TOKENIZER_H_
#define _TOKENIZER_H_

#include <cstdio>
#include <iostream>
#include <string>

// forward declarations
class Tokenizer;
std::ostream& operator<<(std::ostream&, const Tokenizer&);

/**
* @class Tokenizer
* @ingroup util
*
* This class parses an input stream into "tokens", which are read one at a
* time.
*
* The parsing process is controlled by a syntax table that recognizes
* words, numbers, symbols, and quoted strings.  Each character read from the
* input stream is classified as <i>white space</i>, <i>ordinary</i> or a
* <i>word character</i>: white space is ignored, ordinary characters are
* parsed as individual tokens, and word characters are parsed in a string.
* By default, characters in the range <code>'00'</code> to <code>'20'</code>
* are white space, and <code>'21'</code> to <code>'FF'</code> are word
* characters, with the exception of the quote character <code>'"'</code>.
*
* A typical application first constructs an instance of this class,
* initializes the syntax table, and enters a loop, calling the
* <code>next()</code> method in each iteration until it evaluates to
* <code>false</code>.  For example:
* <pre>
*   Tokenizer token(cin);
*   while (token.next()) {
*       if (token == "end")
*           break;
*       else
*           cout << token << endl;
*   }
* </pre>
*/
class Tokenizer {
    // friends
    friend std::ostream& operator<<(std::ostream&, const Tokenizer&);

private:  // types
    enum { TOKEN_LENGTH = 1023 };

public:  // methods
    /**
    * Constructs a tokenizer that parses the specified input stream.
    * @param in  input stream
    */
    Tokenizer(std::istream& in);

    /**
    * Constructs a tokenizer that parses the specified character string.
    * @param string  character string
    */
    Tokenizer(const std::string& string);

    /**
    * Destructor.
    */
    ~Tokenizer();

    /**
    * Returns true if the current token is identical to the specified
    * symbol.
    * @param symbol  symbol
    * @return true if the current token is identical to <i>symbol</i>
    */
    bool operator==(char symbol) const;

    /**
    * Returns true if the current token is identical to the specified
    * string.
    * @param string  string
    * @return true if the current token is identical to <i>string</i>
    */
    bool operator==(const char* string) const;

    /**
    * Returns true if the current token is different from the specified
    * symbol.
    * @param symbol  symbol
    * @return true if the current token is different from <i>symbol</i>
    */
    bool operator!=(char symbol) const;

    /**
    * Returns true if the current token is different from the specified
    * string.
    * @param string  string
    * @return true if the current token is different from <i>string</i>
    */
    bool operator!=(const char* string) const;

    /**
    * Returns true if the end of the stream is reached.
    * @return true if end of stream is reached
    */
    bool operator!() const;

    /**
    * Returns true if a token is available.  Returns false if the end
    * of the stream is reached.
    * @return true if a token is available
    */
    operator bool() const;

    /**
    * Returns the current token.
    * @return current token
    */
    operator const char*() const;

    /**
    * Returns the current line number.
    * @return current line number
    */
    int line() const;

    /**
    * Parses the next token in the stream.
    * @return this tokenizer
    */
    Tokenizer& next();

    /**
    * Parses the remainder of the line as one token.  Leading and trailing
    * whitespace is ignored.
    * @return this tokenizer
    */
    Tokenizer& nextLine();

    /**
    * Defines the specified character as ordinary.
    * @param ordinary  ordinary character
    */
    void ordinaryChar(unsigned char ordinary);

    /**
    * Defines the characters of the specified string as ordinary.
    * @param ordinary  string of ordinary characters
    */
    void ordinaryChars(const char* ordinary);

    /**
    * Defines the characters in the specified range as ordinary.
    * @param first  first ordinary character
    * @param last  last ordinary character
    */
    void ordinaryChars(unsigned char first, unsigned char last);

    /**
    * Causes the next call to <code>next()</code> to return the current token.
    * @see next()
    */
    void putback();

    /**
    * Returns the quote character for the current token.  If the token is not
    * a quoted string, then <code>0</code> is returned.
    * @return quote character for the current token
    */
    char quote() const;

    /**
    * Defines the specified character as a quote.
    * @param quote  quote character
    */
    void quoteChar(unsigned char quote);

    /**
    * Resets the syntax to the default.
    */
    void resetSyntax();

    /**
    * Converts the current token to its numeric value.
    * @param value  address to store value
    * @return false if current token is not a number
    */
    bool value(double* value) const;

    /**
    * Converts the current token to its numeric value.
    * @param value  address to store value
    * @return false if current token is not a number
    */
    bool value(float* value) const;

    /**
    * Converts the current token to its numeric value.
    * @param value  address to store value
    * @return false if current token is not a number
    */
    bool value(int* value) const;

    /**
    * Defines the specified character as whitespace.
    * @param whitespace  whitespace character
    */
    void whitespaceChar(unsigned char whitespace);

    /**
    * Defines the characters in the specified string as whitespace.
    * @param whitespace  string of whitespace characters
    */
    void whitespaceChars(const char* whitespace);

    /**
    * Defines the characters in the specified range as whitespace.
    * @param first  first whitespace character
    * @param last  last whitespace character
    */
    void whitespaceChars(unsigned char first, unsigned char last);

    /**
    * Defines the specified character as a word character.
    * @param word  word character
    */
    void wordChar(unsigned char word);

    /**
    * Defines the characters of the specified string as word characters.
    * @param word  string of word characters
    */
    void wordChars(const char* word);

    /**
    * Defines the characters in the specified range as word characters.
    * @param first  first word character
    * @param last  last word character
    */
    void wordChars(unsigned char first, unsigned char last);

private:  // methods
    Tokenizer();

    Tokenizer(const Tokenizer&);

    void operator=(const Tokenizer&);

    bool _isOrdinary(int c) {
        return (_type[c] == ORDINARY);
    }

    bool _isQuote(int c) {
        return (_type[c] == QUOTE);
    }

    bool _isWhitespace(int c) {
        return (_type[c] == WHITESPACE);
    }

    bool _isWord(int c) {
        return (_type[c] == WORD);
    }

private:  // state
    /** The input stream. */
    std::istream* _in;

    /** The current line number. */
    int _line;

    /** Delete the input stream if true. */
    bool _owner;

    /** Repeat current token if true. */
    bool _putback;

    /** The quote symbol for the current token. */
    char _quote;

    /** The current token. */
    char _token[TOKEN_LENGTH + 1];

    /** The syntax table. */
    enum { ORDINARY, QUOTE, WHITESPACE, WORD } _type[256];
};

inline bool Tokenizer::operator==(char symbol) const {
    return (*_token == symbol && *(_token + 1) == 0);
}

inline bool Tokenizer::operator==(const char* string) const {
    return (strcmp(string, _token) == 0);
}

inline bool Tokenizer::operator!=(char symbol) const {
    return (!(*this == symbol));
}

inline bool Tokenizer::operator!=(const char* string) const {
    return (!(*this == string));
}

inline bool Tokenizer::operator!() const {
    return (*_token == 0);
}

inline Tokenizer::operator bool() const {
    return (*_token != 0);
}

inline Tokenizer::operator const char*() const {
    return (_token);
}

inline int Tokenizer::line() const {
    return (_line);
}

inline void Tokenizer::putback() {
    if (*_token != 0)
        _putback = true;
}

inline char Tokenizer::quote() const {
    return (_quote);
}

inline bool Tokenizer::value(double* value) const {
    return (sscanf(_token, "%lf", value) == 1);
}

inline bool Tokenizer::value(float* value) const {
    return (sscanf(_token, "%f", value) == 1);
}

inline bool Tokenizer::value(int* value) const {
    return (sscanf(_token, "%d", value) == 1);
}

#endif  // _TOKENIZER_H_
